(*** LOCKRES Version 2.3.
Type checking algorithm for session types in the pi calculus
which resolves a sub-class of wait-for deadlocks by generating
a program transformation. The theory has been presented at 
Trusworthly Global Computing (TGC), Universidad de 
Buenos Aires, August 2013 [6]. For a gentle introduction to 
session  types see the interview with Steve Ross-Talbot [5]. 
Main maintainer: Marco Giunti, ctp.di.fct.unl.pt/~mgiunti.
Web: ctp.di.fct.unl.pt/~mgiunti/lockres
Top-level call: fun typeCheckAndResolve
***)

(*** The general syntax of types include input/output session types qualified as linear or 
unrestricted. 
%
NOTE: in the top level call we enforce usage of linear qualifiers, since
the deadlock resolution part of the algorithm currently works with linear session types.
The datatype definitions are kept for backward compatibility and future extensions, which 
include unrestricted qualifiers.
%
A linear type is used exactly once, possibly multiple times (because of type evolution).
A session end point  In (a,c) describes the input of an argument of type a and then
the evolution to type c. A end point type Out (a,c) describes the output of an arguments
of type a and then the evolution as c. The session end point End describes the end of the
session. Channel types have the form Channel (s1,s2) where each s_i is a session end point
or the special type Void, which is used in the type checking algorithm to forbid the use
of a consumed linear end point. 
***) 


datatype qualifier  = Lin | Un;
datatype  preType = Start | End 
                    | In  of  sessionType *  endpointType  
                    | Out of  sessionType *  endpointType
and  endpointType = Qualified of qualifier *  preType 
                    | Void
and  sessionType  = Channel of  endpointType *  endpointType
(*                     | Projection of endpointType *  endpointType *)
                    | EndPoint of  endpointType
                    | Base;
type var           = string; 
type context       = (var * sessionType) list; 
  
(*****************************Pi calculus processes***************************) 
(***Polymorphic implementation of pi calculus process. Channels carry channels 
and values of base type ***) 
(*** Processes with decorated channels - used by Versions 2.x, disentangling***)


datatype 'a dprocess = DContinuation | DZero
                    | DReplication of 'a dprocess 
                    | DParallel of 'a dprocess * 'a dprocess 
                    | DInput of  (string * endpointType * bool) * string * 'a dprocess 
                    | DOutput of (string * endpointType * bool) * 'a dvalue * 'a dprocess 
                    | DNew of string *  sessionType * 'a dprocess 
                    | DIfThenElse of 'a dvalue * 'a dprocess * 'a dprocess
and      'a dvalue   = DValue of 'a | DVar of string; 

(******************Exception declarations************************************)

exception typeFail;
exception contextFail;
exception contextFailL;
exception sbalFail;
exception notConsumed;
exception checkVarFail;   
exception checkFail; 
exception projectionFail;
exception decorationFail;
exception swapFail;
exception disentangleFail;
exception inputContextWrong;
exception innerDisentangleFail;
 
 
(*****************************Functions over session types*********************)


(*** Setting a end point type to void ***)
fun setVoid (EndPoint s) = Void
 |  setVoid (_) = raise typeFail;



(* Casting of two end point types to a channel type *)
fun upCast (EndPoint e, EndPoint f)= Channel (e,f)
 |  upCast (_,_) = raise typeFail;

(* Casting of a session type to an end point type *)
fun downCast ( EndPoint e)= e
 |  downCast (_) = raise typeFail
 
(*** Defining the dual end point types. Used in fun disgt ***)
fun dual ( Qualified  (Un,End))      = Qualified  (Un,End) 
 |  dual ( Qualified  (q,In (a,c))) = Qualified  (q,Out (a,dual c))  
 |  dual ( Qualified  (q,Out (a,c)))= Qualified  (q,In (a,dual c)) 
 |  dual(_) = raise sbalFail;
 
(*** Defining balanced types. Used in fun sbal ***)
fun balanced (Qualified  (Un,End), Qualified  (Un,End))=true  
 |  balanced (Qualified  (p,In (a,c)), Qualified  (q,Out (b,d)) ) =
     p=q andalso a=b andalso balanced (c,d)
 |  balanced (Qualified  (p,Out (a,c)), Qualified  (q,In (b,d)) ) =
     p=q andalso a=b andalso balanced (c,d)
 |  balanced(_,_) = raise sbalFail;
     
 
(*** Defining the class of strongly balanced (linear) types. 
Used in fun sbalContext  and in the restriction case in fun check ***)
fun sbal (Base) = Base
 |  sbal (EndPoint (Qualified  (Un,End))) = EndPoint (Qualified  (Un,End))
 |  sbal (EndPoint (Qualified  (Lin,In (a,c)))) =
      let val t = downCast (sbal ( EndPoint (c) ) ) 
       in EndPoint (Qualified  (Lin,In (sbal(a),t) ) )
      end
 |  sbal (EndPoint (Qualified  (Lin,Out (a,c)))) =
      let val t = downCast (sbal ( EndPoint (c) ) ) 
       in EndPoint (Qualified  (Lin,Out (sbal(a),t) ) )
      end
 |  sbal (Channel (e,g)) = 
      let val s = sbal (EndPoint e) in
       let val t = sbal (EndPoint g) in
        if balanced (downCast (s),downCast(t) ) then
         upCast (  s ,  t )
        else
         raise sbalFail
       end
      end
 | sbal(_) = raise sbalFail;
  

(* Checking if a type is unrestricted or void. In case, returns the session type
received in input. Uses fun upCast. Used in fun consumed. *)
fun unRestricted (Base) = Base
 |  unRestricted (EndPoint Void) = EndPoint Void
 |  unRestricted (EndPoint (Qualified (Un,p))) = EndPoint (Qualified (Un,p))  
 |  unRestricted (Channel  (e,f)) =  
     let val s = unRestricted (EndPoint e) 
      in let val t= unRestricted (EndPoint f)
       in upCast (s,t)
      end
     end
 | unRestricted(_) = raise notConsumed;
     
     
(*Composing two types. The return type to is set to void if the two types are 
equal and linear. Used in fun checkVar.*)

fun compType (Base,Base) = Base
 |  compType (EndPoint (Void),EndPoint (Void)) = EndPoint (Void) 
 |  compType (EndPoint (Qualified (Un,p)),EndPoint (Qualified (Un,r))) =
     if p=r then
      EndPoint (Qualified (Un,p))
     else
      raise typeFail
 |  compType (EndPoint (Qualified (Lin,p)),EndPoint (Qualified (Lin,r))) =
     if p=r then
      EndPoint (Void)
     else
      raise typeFail
 |  compType (Channel  ( Qualified (Un,p), Qualified (Un,s) ),  
               Channel  ( Qualified (Un,r), Qualified (Un,t) )) =
     if p=r andalso s=t then 
          Channel  ( Qualified (Un,p), Qualified (Un,s))  
     else
      if p=t andalso s=r then
          Channel  ( Qualified (Un,p), Qualified (Un,s)) 
      else
       raise typeFail
 |  compType (Channel  ( Qualified (Lin,p), Qualified (Lin,s) ),  
               Channel  ( Qualified (Lin,r), Qualified (Lin,t) )) =
     if p=r andalso s=t then 
          Channel  ( Void, Void ) 
     else
      if p=t andalso s=r then
          Channel  ( Void, Void )
      else
       raise typeFail
 | compType (Channel   ( Qualified (Un,p),  e), EndPoint  ( Qualified (Un,r)))=
     if p=r   then 
       (Channel   ( Qualified (Un,p),  e))  
     else 
       let val s = compType (EndPoint (e), EndPoint ( Qualified (Un,r)))
        in let val t=downCast(s) 
         in Channel ( Qualified (Un,p),t )
        end
       end
 | compType (Channel   ( e,  Qualified (Un,p)), EndPoint  ( Qualified (Un,r)))=
     if p=r   then 
       Channel ( e,Qualified (Un,p) )  
     else 
       let val s = compType (EndPoint (e), EndPoint ( Qualified (Un,r)))
        in let val t=downCast(s) 
         in Channel ( t, Qualified (Un,p))
        end
       end
 | compType (Channel   ( Qualified (Lin,p),  e), EndPoint  ( Qualified (Lin,r)))=
     if p=r   then 
       Channel ( Void,e )  
     else 
       let val s = compType (EndPoint (e), EndPoint ( Qualified (Lin,r)))
        in let val t=downCast(s) 
         in Channel ( Qualified (Lin,p),t )
        end
       end
  | compType (Channel   ( e,  Qualified (Lin,p)), EndPoint  ( Qualified (Lin,r)))=
     if p=r   then 
       Channel ( e,Void )  
     else 
       let val s = compType (EndPoint (e), EndPoint ( Qualified (Lin,r)))
        in let val t=downCast(s) 
         in Channel ( t, Qualified (Lin,p))
        end
       end 
  | compType (_,_)= raise checkVarFail;
   
  
(*** Creation of a linear session type of depth equal to one***)
(*** Used in Versions 2.x, fun disgt ***)

fun ioType (t:sessionType) = 
  Channel (Qualified (Lin,(In (t,Qualified (Un,End)))), 
           Qualified (Lin,(Out (t,Qualified (Un,End))))); 
  
(*** Macro for CCS channels ***)
(*** Used in Versions 2.x, fun disgt ***)

val voidType =  Channel (Void,Void);
val vType = voidType;
val ioVoid = ioType(voidType); 

 

(***************************Functions over contexts****************************) 

(*** Put on top of the stack a variable ***)

fun topStack (nil:context,v:var):context= raise contextFail
|   topStack ((x,s)::g,y) =
     if x=y then 
      (x,s)::g
     else
      let val d = topStack (g,y)
       in d@[(x,s)]
      end


(***Removing an entry from the context***)
fun remove (nil:context,v:var):context = raise contextFail
 |  remove ((x,t)::g,v)  =
     if x=v then g
     else 
      let val d = remove(g,v)
       in (x,t)::d
      end;  
      
(***Returning the type of a variable***)      
fun isType(nil:context,v:var):sessionType = raise contextFail
 | isType((x,s)::g,v) = 
     if x=v then s
     else
      isType(topStack ((x,s)::g,v),v);
      
     
(*** Checking if the the type associated to a variable is
unrestricted or void. Used in rules for input and output in fun check***) 
fun consumed (nil:context, v:var):context =  raise contextFail
 |  consumed ((x,s)::g,v) = 
     if x=v then 
      let val t= unRestricted (s) 
       in
         (x,t)::g 
      end
     else
      let val d = consumed (g,v)
       in (x,s)::d
      end;    
      
(*** Checking if the environment has been all consumed ***)
fun finished (nil)=nil
 |  finished ((x,s)::g) = 
     let val t= unRestricted (s) 
       in (x,t):: finished (g)
     end

(*** Setting a end point type in the domain of a context to void ***)
fun setVoid (nil:context,v:var):context = raise contextFail
 |  setVoid ((x,EndPoint s)::g,v:var) =
      if x=v then 
       (x,EndPoint Void)::g
      else
       let val d = setVoid(g,v)
        in (x,EndPoint s)::d
       end
 |  setVoid (_,_) = raise typeFail;

(*** Transforming an end point type in the domain of a context in a channel type ***)
fun upCastContext (nil:context,v:var, EndPoint e):context = raise contextFail
 |  upCastContext ((x,EndPoint h)::g, v, EndPoint e) =
     if x=v then
      let val s = upCast (EndPoint h,EndPoint e) 
       in (x,s)::g
      end
     else
      let val d = topStack ( (x,EndPoint h)::g,v)
       in  upCastContext (d,v,EndPoint e)
      end
 | upCastContext ((x,Channel (h1,h2))::g, v, EndPoint e) =
     if x=v then
      raise contextFail
     else
      let val d = topStack ( (x,Channel (h1,h2))::g,v)
       in  upCastContext (d,v,EndPoint e)
      end
 | upCastContext ((x,Base)::g, v, EndPoint e) =
     if x=v then
      raise contextFail
     else
      let val d = topStack ( (x,Base)::g,v)
       in  upCastContext (d,v,EndPoint e)
      end
 | upCastContext(_,_,_)= raise contextFail;
 
 fun upCastContextL (nil:context,v:var, EndPoint e):context = raise contextFailL
 |  upCastContextL ((x,EndPoint h)::g, v, EndPoint e) =
     if x=v then
      let val s = upCast (EndPoint e,EndPoint h) 
       in (x,s)::g
      end
     else
      let val d = topStack ( (x,EndPoint h)::g,v)
       in  upCastContextL (d,v,EndPoint e)
      end
 | upCastContextL ((x,Channel (h1,h2))::g, v, EndPoint e) =
     if x=v then
      raise contextFailL
     else
      let val d = topStack ( (x,Channel (h1,h2))::g,v)
       in  upCastContextL (d,v,EndPoint e)
      end
 | upCastContextL ((x,Base)::g, v, EndPoint e) =
     if x=v then
      raise contextFailL
     else
      let val d = topStack ( (x,Base)::g,v)
       in  upCastContextL (d,v,EndPoint e)
      end
 | upCastContextL(_,_,_)= raise contextFailL;
 
(*** Checking if an environment is strongly balanced.
Used in the top-level call fun typeCheckAndResolve.
Used to enforce that in delegation both endpoints of a channel
are delegated. This is a limitation that we plan to solve in 
future releases.
***)

fun sbalContext (nil)=nil
 |  sbalContext ((x,s)::g) =
      (x,sbal(s))::sbalContext (g); 

(*** Creating a projection in the top level call of disentangling ***)
(*** Used by Versions 2.x, fun typeCheckAndResolve ***)

fun projection (nil)= nil
 |  projection ((x,s)::g) = 
     (x,EndPoint (Qualified (Un,Start)))::projection(g);
     
fun swap ((x,s)::l,(y,t)::m,z) =
  if x=y andalso y=z then
   ((x,t)::l,(y,s)::m)
  else
   swap(topStack((x,s)::l,z),topStack((y,t)::l,z),z)
 | swap(_,_,_)= raise swapFail;
                
      
(**************************Type Checking************************************)

(* Casting a string in output *)
fun stringCast (DVar (s))=s
 |  stringCast(_) = raise typeFail;

 
(*** Type checking a variable. Used in the patterns for output in fun check***)

fun checkVar( nil:context, (z,t):(var * sessionType) ):context = raise checkVarFail 
 |  checkVar( (x,s)::g, (z,t) ) = 
     if x=z then
      let val u = compType(s,t) in
       (x,u)::g
      end
     else
      let val d=checkVar (g,(z,t)) in
        (x,s)::d
      end;
  
(************************Type checking and disentangling process**********************)
(*****************************Version 2.2, November 2012***************************)

fun dcheckVar( nil,d, (z,t)) = raise checkVarFail 
 |  dcheckVar( (x,s)::g,d, (z,t)) = 
     if x=z then
      let val u = compType(s,t) in
       ((x,u)::g,d)
      end
     else
      let val (g1,d1)=dcheckVar (g,d,(z,t)) in
        ((x,s)::g1,d1)
      end;


fun fdisgt 
(* (DContinuation ,(x,s),y,v)=(DContinuation,v) *)
  (DZero,(x,s),y,v)=(DZero,v)
(*  |  fdisgt (h,(x,Qualified (Un,End)),y,v)= (h,v) *)
 |  fdisgt (DReplication h,(x,s),y,v)= (DReplication h,v)
 |  fdisgt (DNew (z,t,h), (x,s),y,v)=
      let val (k,v)= fdisgt (h,(x,s),y,v) in
       (DNew (z,t,k),v) 
      end
 |  fdisgt (DParallel (h1,h2),(x,s),y,v) = 
     let val (k1,v1)=fdisgt (h1,(x,s),y,v) 
         val (k2,v2)=fdisgt (h2,(x,s),y,v) 
      in
       if v1=v then (DParallel (k1,k2),v2)
       else         (DParallel (k1,k2),v1) 
     end
 |  fdisgt (DIfThenElse (w,h1,h2),(x,s),y,v)=
     let val (k1,v1)=fdisgt (h1,(x,s),y,v) 
         val (k2,v2)=fdisgt (h2,(x,s),y,v) 
      in
       if v1=v then (DIfThenElse (w,k1,k2),v2)
       else         (DIfThenElse (w,k1,k2),v1) 
     end
 |  fdisgt (DInput ((z,s,f), y,h),(x,t),r,v) = 
      if x=z andalso s=t then
       (DInput ((r,Void,false), y,h),v)
      else
       let val (k,w)=fdisgt (h,(x,t),r,v) in
        (DInput ((z,s,f), y,k),w)
       end 
 |  fdisgt (DOutput ((z,s,f),w,h),(x,t),m,v) =
      if x=z andalso s=t then
       (DOutput ((m,Void,false), DVar "emptychan" ,h),w)
      else
       let val (k,w1)=fdisgt (h,(x,t),m,v) in
        (DOutput ((z,s,f), w,k),w1)
       end
 | fdisgt (_,_,_,_) = raise innerDisentangleFail;
       

fun disgt2 (DZero)=DZero
 |  disgt2 (DReplication h)= DReplication h  
 |  disgt2 (DParallel (h,k)) = 
     DParallel (disgt2 (h),disgt2 (k))
 |  disgt2 (DNew (y,t,h))=
      DNew (y,t,disgt2 (h))
 |  disgt2 (DIfThenElse (y,h,k))=
      DIfThenElse (y,disgt2 (h),disgt2 (k))
 |  disgt2 (DOutput ((x,Qualified (Lin, Out (a,c)),true),y,h)) = 
      let val s        = Qualified (Lin, Out (a,c))
          val ds       = Qualified (Lin, In  (a,c))
          val forwChan =  ("r"^x)^(stringCast y)
          val forwType = ioType (a) 
          val (k,w)    =  fdisgt(h,(x,dual s),forwChan, y)
                         
        in 
          DNew(forwChan,forwType,DParallel 
                        (DOutput ((x,s,false), w, disgt2 (k) ),
                         DInput ((x,ds,false),"z",
                             DOutput ((forwChan,Void,false),DVar "z",DZero))
                         )) 
      end 
 |  disgt2 (DOutput ((x,s,false),y,h)) =  
        DOutput ((x,s,false),y,disgt2(h)) 
 |  disgt2 (DInput ((x,Qualified (Lin, In (a,c)),true),y,h)) =
      let val s        = Qualified (Lin, In (a,c))
          val forwChan = ("r"^x)^y
          val ccsChan  = ("m"^x)^y
          val witness  = ("t"^x)^y
          val forwType = ioType (a)
          val ccsType  = ioVoid 
          val (k,w)    = fdisgt(h,(x,dual s),ccsChan, DVar witness)
       in
          DNew(forwChan,forwType, DNew(ccsChan,ccsType, 
            DParallel (
             DInput ((x,s,false), "z", DOutput ((forwChan,Void,false),DVar "z",DZero)),
             DParallel (
               DInput ((forwChan,Void,false),y, disgt2 (k) ),
               DOutput ((x,dual s,false), w, 
                  DInput ((ccsChan,Void,false),"emptyVar",DZero))
                       )    
                      )
                ))
      end 
 |  disgt2 (DInput ((x,s,false),y,h)) =  
        DInput ((x,s,false),y,disgt2(h))
 | disgt2(_)= raise disentangleFail;


fun disCheck2 (g:context,d:context,DZero)  = (g,d,DZero)  
(* A-REPL   *)
 |  disCheck2 (g,d,DReplication k)         =                  
      let val (f,r,h) = disCheck2(g,d,k) 
       in if f=g andalso r=d then (f,r,DReplication h)
          else raise checkFail
      end 
(* A-PAR *)  
  |  disCheck2 (g,d,DParallel (k1,k2))      =                 
      let val (f1,d1,h1) = disCheck2(g,d,k1)  
          val (f2,d2,h2) = disCheck2(f1,d1,k2) 
        in
          (f2,d2, DParallel (h1,h2)) 
      end        
 (* A-OutInit-L*)
  |  disCheck2 ((z,(Channel ( Qualified (Lin,Out(a,c)),e)))::g,
                (w,(EndPoint (Qualified (Un,Start))))::d1, 
                 DOutput ((x,s,f),DVar (y),p) ) =
    let val s1 = Qualified (Lin,Out(a,c)) 
        val s2 = Qualified (Lin,In(a,c)) 
        val  r = Qualified (Un,Start)
        val  u = Qualified (Un,End) in 
      if z=x andalso w=x then
       let val g1        = checkVar (g,(stringCast (DVar (y)),a))  
           val (g2,d2,h) = disCheck2 ((z,EndPoint c)::g1,(z,(EndPoint e))::d1, p)  
         in 
           if (EndPoint e)=isType(d2,x) 
             then
             (upCastContext (consumed(g2,z),z,EndPoint e),
              (x,EndPoint r)::remove(d2,x),
              DOutput ((x,s1,false),DVar y,h))
            else 
               (upCastContext (consumed(g2,z),z,EndPoint u),
               (x,EndPoint r)::remove(consumed(d2,x),x),
               disgt2( DOutput((x,s1,true),DVar (y),h) )
               ) 
       end
      else
       if w=x then 
        let  val g1=topStack((z,(Channel ( s1,e)))::g,x) in
          disCheck2(g1,(w,EndPoint r)::d1,DOutput ((x,s,f),DVar y,p))
        end
       else
        if z=x then 
        let
            val r1=topStack((w,EndPoint r)::d1,x) 
          in
            disCheck2((z,(Channel ( s1,e)))::g,r1,DOutput ((x,s,f),DVar y,p))  
        end
        else
         let  
           val g1=topStack((z,(Channel ( s1,e)))::g,x)
           val r1=topStack((w,EndPoint r)::d1,x) in
             disCheck2(g1,r1,DOutput ((x,s,f),DVar y,p))
         end
    end
  (* A-OutInit-R*)
  |  disCheck2 ((z,(Channel ( e,Qualified (Lin,Out(a,c)))))::g,
                (w,(EndPoint (Qualified (Un,Start))))::d1, 
                 DOutput ((x,s,f),DVar (y),p) ) =
     let val s1 = Qualified (Lin,Out(a,c)) 
        val s2 = Qualified (Lin,In(a,c)) 
        val  r = Qualified (Un,Start)
        val  u = Qualified (Un,End) in 
      if z=x andalso w=x then
       let val g1        = checkVar (g,(stringCast (DVar (y)),a))  
           val (g2,d2,h) = disCheck2 ((z,EndPoint c)::g1,(z,(EndPoint e))::d1, p)  
         in 
           if (EndPoint e)=isType(d2,x) 
             then
             (upCastContextL (consumed(g2,x),x,EndPoint e),
              (x,EndPoint r)::remove(d2,x),
              DOutput ((x,s1,false),DVar y,h))
            else 
               (upCastContextL (consumed(g2,x),x,EndPoint u),
               (x,EndPoint r)::remove(consumed(d2,x),x),
               disgt2( DOutput((x,s1,true),DVar (y),h) )
               ) 
       end
      else
       if w=x then 
        let  val g1=topStack((z,(Channel ( e,s1)))::g,x) in
          disCheck2(g1,(w,EndPoint r)::d1,DOutput ((x,s,f),DVar y,p))
        end
       else
        if z=x then 
        let
            val r1=topStack((w,EndPoint r)::d1,x) 
          in
            disCheck2((z,(Channel ( e,s1)))::g,r1,DOutput ((x,s,f),DVar y,p))  
        end
        else
         let  
           val g1=topStack((z,(Channel ( e,s1)))::g,x)
           val r1=topStack((w,EndPoint r)::d1,x) in
             disCheck2(g1,r1,DOutput ((x,s,f),DVar y,p))
         end
    end   
  (* A-Out-Handled*)
  |  disCheck2 ((z,EndPoint ( Qualified (Lin,Out(a,c))))::g,
                (w,EndPoint ( Qualified (Lin,Out(b,e))))::d, 
                DOutput ((x,s,f),DVar (y),p) ) =
     let val s1 = Qualified (Lin,Out(a,c))
         val s2 = Qualified (Lin,Out(b,e)) in
      if z=x andalso w=x then 
       let
           val (g1,d1)   = dcheckVar (g,d,(stringCast (DVar (y)),a)) 
                           handle checkVarFail => 
                             dcheckVar (g,d,(stringCast (DVar(y)),b))  
           val (g2,d2,h) = 
                  disCheck2 ((z,EndPoint c)::g1, (z,EndPoint s2)::("fg",Base)::d1,p) 
                  handle checkFail => 
                     disCheck2 ((z,EndPoint e)::g1, (z,EndPoint s1)::("fg",vType)::d1,p) 
         in
          let val (g3,d3)=swap(g2,remove (d2,"fg"),z) in
           if Base=isType(d2,"fg") 
           then 
            if (EndPoint s2)=isType(d2,x) 
             then
             (g3,d3,DOutput ((x,s1,false),DVar y,h)) 
            else 
             (g3,d3,DOutput ((x,s1,true),DVar (y),h)) 
           else
            if (EndPoint s1)=isType(d2,x) 
             then
             (g3,d3,DOutput ((x,s2,false),DVar y,h)) 
            else 
             (g3,d3,DOutput ((x,s2,true),DVar (y),h)) 
          end
       end
      else
       if w=x then 
        let  val g1=topStack((z,EndPoint s1)::g,x) in
          disCheck2(g1,(w,EndPoint s2)::d,DOutput ((x,s,f),DVar y,p))
        end
       else
        if z=x then 
        let
            val r1=topStack((w,EndPoint s2)::d,x) 
          in
            disCheck2((z,EndPoint s1)::g,r1,DOutput ((x,s,f),DVar y,p))  
        end  
       else
       let  val g1=topStack((z,EndPoint s1)::g,x)  
            val r=topStack((w,EndPoint s2)::d,x) 
          in
            disCheck2(g1,r,DOutput ((x,s,f),DVar y,p))  
       end
     end
    (* A-OutEnv*)
  |  disCheck2 ((z,EndPoint ( Qualified (Lin,Out(a,c))))::g,
                (w,EndPoint e)::d, 
                DOutput ((x,s,f),DVar (y),p) ) =
     let val s1 = Qualified (Lin,Out(a,c)) in
      if z=x andalso w=x then 
       let
           val (g1,d1)   = dcheckVar (g,d,(stringCast (DVar (y)),a))   
           val (g2,d2,h) = disCheck2 ((z,EndPoint c)::g1, (w,EndPoint e)::d1,p)   
         in 
            if (EndPoint e)=isType(d2,x) 
             then
             (g2,d2,DOutput ((x,s1,false),DVar y,h)) 
            else 
             (g2,d2,DOutput ((x,s1,true),DVar (y),h))  
       end
      else 
       if w=x then 
        let  val g1=topStack((z,(EndPoint s1))::g,x) in
          disCheck2(g1,(w,EndPoint e)::d,DOutput ((x,s,f),DVar y,p))
        end
       else
        if z=x then 
        let
            val r1=topStack((w,EndPoint e)::d,x) 
          in
            disCheck2((z,EndPoint s1)::g,r1,DOutput ((x,s,f),DVar y,p))  
        end
       else
       let  val g1=topStack((z,EndPoint s1)::g,x)  
            val r=topStack((w,EndPoint e)::d,x) 
          in
            disCheck2(g1,r,DOutput ((x,s,f),DVar y,p))  
       end
     end 
     (* A-OutProj*)
  |  disCheck2 ((z,EndPoint e)::g,
                (w,EndPoint ( Qualified (Lin,Out(a,c))))::d, 
                 DOutput ((x,s,f),DVar (y),p) ) =
     let val s1 = Qualified (Lin,Out(a,c)) in
      if z=x andalso w=x then 
       let
           val (g1,d1)   = dcheckVar (g,d,(stringCast (DVar (y)),a))   
           val (g2,d2,h) = disCheck2 ((z,EndPoint e)::g1, (w,EndPoint c)::d1,p)   
         in 
            if (EndPoint e)=isType(g2,x) 
             then
             (g2,d2,DOutput ((x,s1,false),DVar y,h)) 
            else 
             (g2,d2,DOutput ((x,s1,true),DVar (y),h))  
       end
      else
       if w=x then 
        let  val g1=topStack((z,(EndPoint e))::g,x) in
          disCheck2(g1,(w,EndPoint s1)::d,DOutput ((x,s,f),DVar y,p))
        end
       else
        if z=x then 
        let
            val r1=topStack((w,EndPoint s1)::d,x) 
          in
            disCheck2((z,EndPoint e)::g,r1,DOutput ((x,s,f),DVar y,p))  
        end 
       else
       let  val g1=topStack((z,EndPoint e)::g,x)  
            val r=topStack((w,EndPoint s1)::d,x) 
          in
            disCheck2(g1,r,DOutput ((x,s,f),DVar y,p))  
       end
     end
(* Else      *)
  | disCheck2 ((z,t)::g, (w,r)::d, DOutput ((x,s,f),DVar (y),p) ) =
      if x=z orelse  w=x then
         raise checkFail
       else
       let  val  g1 = topStack((z,t)::g,x)  
            val  d1 = topStack((w,r)::d,x) 
          in
            disCheck2(g1,d1,DOutput ((x,s,f),DVar y,p))  
       end 
(* A-InInit-L*)
  |  disCheck2 ((z,(Channel ( Qualified (Lin,In(a,c)),e)))::g,
                (w,(EndPoint (Qualified (Un,Start))))::d1, 
                 DInput ((x,s,f),y,p) ) =
    let val s1 = Qualified (Lin,In(a,c)) 
        val s2 = Qualified (Lin,Out(a,c)) 
        val  r = Qualified (Un,Start)
        val  u = Qualified (Un,End) in 
      if z=x andalso w=x then
       let val (g2,d2,h) = 
           disCheck2 ((y,a)::(z,EndPoint c)::g,(y,EndPoint r)::(z,(EndPoint e))::d1,p) 
           val g3 = remove (consumed ( consumed(g2,z),y ), y)
           val d3 = remove (consumed (d2,y) , y)
         in 
           if (EndPoint e)=isType(d3,x) 
             then
             (upCastContext ( g3,z,EndPoint e),
              (x,EndPoint r)::remove (d3,w),
              DInput ((x,s1,false),y,h))
            else 
               (upCastContext (g3,z,EndPoint u),
               (x,EndPoint r)::remove(d3,x),
               disgt2( DInput((x,s1,true),y,h) )
(*                disgt2( setDualTrue ( DInput((x,s1),DVar (y),h) , (x,s1) ) ) *)
(*                DInput ((x,s1,true),DVar (y),h)   *)
               ) 
       end
      else
       if w=x then 
        let  val g1=topStack((z,(Channel ( s1,e)))::g,x) in
          disCheck2(g1,(w,EndPoint r)::d1,DInput ((x,s,f),y,p))
        end
       else
        if z=x then 
        let
            val r1=topStack((w,EndPoint r)::d1,x) 
          in
            disCheck2((z,(Channel ( s1,e)))::g,r1,DInput ((x,s,f),y,p))  
        end 
      else
       let  val g1=topStack((z,(Channel ( s1,e)))::g,x)  
            val r1=topStack((w,EndPoint r)::d1,x) 
          in
            disCheck2(g1,r1,DInput ((x,s,f),y,p))  
       end
    end
 (* A-InInit-R*)
  |  disCheck2 ((z,Channel ( e,Qualified (Lin,In(a,c))))::g,
                (w,(EndPoint (Qualified (Un,Start))))::d1, 
                 DInput ((x,s,f),y,p) ) =
    let val s1 = Qualified (Lin,In(a,c)) 
        val s2 = Qualified (Lin,Out(a,c)) 
        val  r = Qualified (Un,Start)
        val  u = Qualified (Un,End) in 
      if z=x andalso w=x then
       let val (g2,d2,h) = 
           disCheck2 ((y,a)::(z,EndPoint c)::g,(y,EndPoint r)::(z,(EndPoint e))::d1,p) 
           val g3 = remove (consumed ( consumed(g2,x),y ), y)
           val d3 = remove (consumed (d2,y) , y)
         in 
           if (EndPoint e)=isType(d3,x) 
             then
             (upCastContextL ( g3,x,EndPoint e),
              (x,EndPoint r)::remove (d3,x),
              DInput ((x,s1,false),y,h))
            else 
               (upCastContextL (g3,x,EndPoint u),
               (x,EndPoint r)::remove(d3,x),
               disgt2( DInput((x,s1,true),y,h) )
(*                disgt2( setDualTrue ( DInput((x,s1),DVar (y),h) , (x,s1) ) ) *)
(*                DInput ((x,s1,true),DVar (y),h)   *)
               ) 
       end
      else
       if w=x then 
        let  val g1=topStack((z,(Channel ( e,s1)))::g,x) in
          disCheck2(g1,(w,EndPoint r)::d1,DInput ((x,s,f),y,p))
        end
       else
        if z=x then 
        let
            val r1=topStack((w,EndPoint r)::d1,x) 
          in
            disCheck2((z,(Channel ( e,s1)))::g,r1,DInput ((x,s,f),y,p))  
        end 
      else
       let  val g1=topStack((z,(Channel ( e,s1)))::g,x)  
            val r1=topStack((w,EndPoint r)::d1,x) 
          in
            disCheck2(g1,r1,DInput ((x,s,f),y,p))  
       end
    end
  (* A-In-Handled*)
  |  disCheck2 ((z,EndPoint ( Qualified (Lin,In(a,c))))::g,
                (w,EndPoint ( Qualified (Lin,In(b,e))))::d, 
                DInput ((x,s,f),y,p) ) =
     let val s1 = Qualified (Lin,In(a,c))
         val s2 = Qualified (Lin,In(b,e)) 
         val start = EndPoint (Qualified (Un,Start)) in
      if z=x andalso w=x then 
       let 
           val (g2,d2,h) = 
                  disCheck2 ((y,a)::(z,EndPoint c)::g, 
                             (y,start)::(z,EndPoint s2)::("fg",Base)::d,
                             p)
                  handle checkVarFail => 
                     disCheck2 ((y,b)::(z,EndPoint e)::g, 
                                (y,start)::(z,EndPoint s1)::("fg",vType)::d,
                                p)
                  handle checkFail => 
                     disCheck2 ((y,b)::(z,EndPoint e)::g, 
                                (y,start)::(z,EndPoint s1)::("fg",vType)::d,
                                p) 
         in
          let val (g3,d3)=swap(g2,remove (d2,"fg"),z) in
           if Base=isType(d2,"fg") 
           then 
            if (EndPoint s2)=isType(d2,x) 
             then
             (g3,d3,DInput ((x,s1,false),y,h)) 
            else 
             (g3,d3,DInput ((x,s1,true),y,h)) 
           else
            if (EndPoint s1)=isType(d2,x) 
             then
             (g3,d3,DInput ((x,s2,false),y,h)) 
            else 
             (g3,d3,DInput ((x,s2,true),y,h)) 
          end
       end
      else
       if w=x then 
        let  val g1=topStack((z,EndPoint s1)::g,x) in
          disCheck2(g1,(w,EndPoint s2)::d,DInput ((x,s,f),y,p))
        end
       else
        if z=x then 
        let
            val d1=topStack((w,EndPoint s2)::d,x) 
          in
            disCheck2((z,EndPoint s1)::g,d1,DInput ((x,s,f),y,p))  
        end  
       else
       let  val g1=topStack((z,EndPoint s1)::g,x)  
            val d1=topStack((w,EndPoint s2)::d,x) 
          in
            disCheck2(g1,d1,DInput ((x,s,f),y,p))  
       end
     end 
   (* A-InEnv*)
  |  disCheck2 ((z,EndPoint ( Qualified (Lin,In(a,c))))::g,
                (w,EndPoint e)::d, 
                DInput ((x,s,f),y,p) ) =
     let val s1 = Qualified (Lin,In(a,c)) in
      if z=x andalso w=x then 
       let 
           val (g2,d2,h) = disCheck2 ((y,a)::(z,EndPoint c)::g, (w,EndPoint e)::d,p)   
         in 
            if (EndPoint e)=isType(d2,x) 
             then
             (g2,d2,DInput ((x,s1,false),y,h)) 
            else 
             (g2,d2,DInput ((x,s1,true),y,h))  
       end
      else 
       if w=x then 
        let  val g1=topStack((z,(EndPoint s1))::g,x) in
          disCheck2(g1,(w,EndPoint e)::d,DInput ((x,s,f),y,p))
        end
       else
        if z=x then 
        let
            val r1=topStack((w,EndPoint e)::d,x) 
          in
            disCheck2((z,EndPoint s1)::g,r1,DInput ((x,s,f),y,p))  
        end
       else
       let  val g1=topStack((z,EndPoint s1)::g,x)  
            val r=topStack((w,EndPoint e)::d,x) 
          in
            disCheck2(g1,r,DInput ((x,s,f),y,p))  
       end
     end
       (* A-InProj*)
  |  disCheck2 ((z,EndPoint e)::g,
                (w,EndPoint ( Qualified (Lin,In(a,c))))::d, 
                 DInput ((x,s,f),y,p) ) =
     let val s1 = Qualified (Lin,In(a,c)) in
      if z=x andalso w=x then 
       let 
           val (g2,d2,h) = disCheck2 ((z,EndPoint e)::g, (w,EndPoint c)::d,p)   
         in 
            if (EndPoint e)=isType(g2,x) 
             then
             (g2,d2,DInput ((x,s1,false),y,h)) 
            else 
             (g2,d2,DInput ((x,s1,true),y,h))  
       end
      else
       if w=x then 
        let  val g1=topStack((z,(EndPoint e))::g,x) in
          disCheck2(g1,(w,EndPoint s1)::d,DInput ((x,s,f),y,p))
        end
       else
        if z=x then 
        let
            val d1=topStack((w,EndPoint s1)::d,x) 
          in
            disCheck2((z,EndPoint e)::g,d1,DInput ((x,s,f),y,p))  
        end 
       else
       let  val g1=topStack((z,EndPoint e)::g,x)  
            val d1=topStack((w,EndPoint s1)::d,x) 
          in
            disCheck2(g1,d1,DInput ((x,s,f),y,p))  
       end
     end
(* Else *)
  | disCheck2 ((z,t)::g, (w,r)::d, DInput ((x,s,f),y,p) ) =
       if x=z orelse  w=x then
         raise checkFail
       else
         let  val  g1 = topStack((z,t)::g,x)  
              val  d1 = topStack((w,r)::d,x) 
          in
             disCheck2(g1,d1,DInput ((x,s,f),y,p))   
          end 
(*  Otherwise       *)
     |   disCheck2(_,_,_) = raise inputContextWrong;
 

(*** Top level call - accepts only strongly balanced contexts 
containing in their range strongly balanced types with linear qualifiers ***) 

fun typeCheckAndResolve (g,p) =
  let val (g1,d1,p1) = disCheck2 (sbalContext (g), projection (g), p)
  in 
     (finished (g1), finished (d1), p1)
  end 
    
    

(***************************** References *******************************

[1] Marco Giunti. A type checking algorithm for qualified session types. 
Automated Specification and Verification of Web Systems (WWV) EPTCS 61 91-114 2011.  
[2] Marco Giunti, Vasco T. Vasconcelos: A Linear Account of Session Types in the
Pi Calculus. Concurrency Theory (CONCUR). LNCS 6269  432-446  2010.
[3] Kohei Honda, Vasco T. Vasconcelos, Makoto Kubo: Language Primitives and Type
Discipline for Structured Communication-Based Programming. Programming Languages and
Systems (ESOP).LNCS 1381 122-138  1998.
[4] Robin Milner, Joachim Parrow, David Walker: A Calculus of Mobile Processes, I and II.
Information and Computation, 100(1) 1-77  1992.
[5] A Conversation with Steve Ross-Talbot. ACM Queue 4(2) 14-23 2006 
[6] Marco Giunti, Antonio Ravara. Towards static deadlock resolution in the pi calculus.

*)

 